/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 */

/*
 * anonymizer.c
 * anonymizes the traces
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#define NEED_DESCRIPTIVE_MESSAGES
#include "messages.h"
#include "tracereader.h"
#include "namehash.h"
#include "pidhash.h"
#include "od_cksum.h"
#include "util.h"
#include <unistd.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/blowfish.h>
#include <errno.h>

#define BF_KEYSIZE 16
#define BF_BLOCKSIZE 8

void printhexstr(const char *prefix, unsigned char *str, int len) {
	int i;

	printf("%s0x", prefix);
	for (i = 0; i < len; i++) {
		printf("%02x", str[i]);
	}
	printf("\n");
}

#define LEN_LOC (sizeof(op_t))
#define INIT_LOC (LEN_LOC + sizeof(msg_len_t))
#define BUF_SIZE 4096
struct buffer {
	char *data;
	char *ptr; /* current pointer */
	int size; /* size of data */
};

static int outfd;
static int do_anonymize = 0;
static int total_bytes_anonymized;


/* end encryption related stuff */
void usage()
{
	fprintf(stderr, "Usage: anonymizer [-c] [-b <block-size>] [-d <key>] -z -f <file-name> -a <list of fields to anonymize -k <keyfile> -o <output file name>\n");
	fprintf(stderr, "       -c specifies that the file is checksummed\n");
      	fprintf(stderr, "       -b <block-size> specifies the block size of the checksum, default 4096\n");
	fprintf(stderr, "       -z decompress using zlib\n");
	fprintf(stderr, "       -d <key> specifies the decryption key\n");
	fprintf(stderr, "      -f specifies the file name\n");
	fprintf(stderr, "       -o <output file> output file for anonymized traces\n");
	fprintf(stderr, "       -a <field> fields to anonymize where <field> could be uid, gid, pid, path, pname, str, or timestamp\n");
        fprintf(stderr, "       -a can be specified once for each anonymizable field\n");
	fprintf(stderr, "       -k <keyfile> output file for generated keys\n");
	fprintf(stderr, "       -y overwrite existing output files\n");
	fprintf(stderr, "Note that order of -c and -z are important\n");
	exit(1);
}
static int dumpkeys(int fd, int keylen)
{
	int i;
	for (i = 0; i < ANON_MAX_FIELDS; i++) {
		if (anon_keys[i].valid) {
			int j;
			if (write(fd, anon_keys[i].name, strlen(anon_keys[i].name)) != strlen(anon_keys[i].name))
				return ERR_IO;
			if (write(fd, ": 0x", 4) != 4)
				return ERR_IO;
			for (j = 0; j < keylen; j++) {
				if (write(fd, byte2hex(anon_keys[i].key[j]), 2) != 2)
					return ERR_IO;
			}
			if (write(fd, "\n", 1) != 1)
				return ERR_IO;
		}
	}
	return ERR_OK;

}

int main(int argc, char *argv[])
{
	int err;
	struct parse_params params;
	int c;
	int index;
	char *outfilename = NULL;
	char *keyfilename = NULL;
	int keyfilefd = -1;
	int i;
	int clobber = 0;
	memset(&params, 0, sizeof(params));
	params.cparams.blk_size = CKSUM_BLOCK_SIZE;
	int randomfd;

	if ((randomfd = open("/dev/random", O_RDONLY)) < 0) {
		fprintf(stderr, "Could not open /dev/random to generate encryption keys\n");
		exit(2);
	}

	while ((c = getopt(argc, argv, "zc:b:f:d:o:a:k:y")) != -1) {
		switch (c) {
			case 'c':
				if (params.n_ops == MAX_OPS) {
					fprintf(stderr, "Too many operations specified\n");
					usage();
				}
				params.ops[params.n_ops++] = OP_CKSUM_MD5;
				if (hmac_make_key(&params.cparams, optarg)) {
					fprintf(stderr, "Invalid key specified\n");
					usage();
				}
				break;
			case 'b':
				params.cparams.blk_size = atoi(optarg);
				break;
			case 'z':
				if (params.n_ops == MAX_OPS) {
					fprintf(stderr, "Too many operations specified\n");
					usage();
				}
				params.ops[params.n_ops++] = OP_UNCOMPRESS;
				break;
			case 'f':
				if (params.n_ops == MAX_OPS) {
					fprintf(stderr, "Too many operations specified\n");
					usage();
				}
				params.ops[params.n_ops++] = OP_FILE;
				params.file_name = optarg;
				break;
			case 'd':
				if (params.n_ops == MAX_OPS) {
					fprintf(stderr, "Too many operations specified\n");
					usage();
				}
				params.ops[params.n_ops++] = OP_DECRYPT;
				if (make_key(&params.dparams, optarg)) {
					fprintf(stderr, "Invalid key specified\n");
					usage();
				}
				break;
			case 'o':
				outfilename = optarg;
				break;
			case '?':
				fprintf(stderr, "Unknown option `-%c'.\n\n", optopt);
				usage();
				break;
			case 'a': {
				for (i = 0; i < ANON_MAX_FIELDS; i++) {
					if (strcmp(anon_keys[i].name, optarg) == 0) {
						memset(anon_keys[i].key, 0xc4, BF_KEYSIZE);
						anon_keys[i].valid = 1;
						do_anonymize = 1;
						break;
					}
				}
				if (i == ANON_MAX_FIELDS) {
					fprintf(stderr, "Unknown field type [%s] for anonymization\n", optarg);
					usage();
				}
			}
				break;
			case 'k':
				keyfilename = optarg;
				break;
			case 'y':
				clobber = 1;
				break;
			default:
				usage();
				break;
		}
	}
	index = optind;
	if (!params.file_name) {
		fprintf(stderr, "File name not specified\n");
		usage();
	}
	if (!outfilename) {
		fprintf(stderr, "output file not specified\n");
		usage();
	}
	if (do_anonymize && !keyfilename) {
		fprintf(stderr, "no key file specified\n");
		usage();
	}

	if (do_anonymize) {
		if ((keyfilefd = open(keyfilename, O_WRONLY | O_CREAT | (!clobber ? O_EXCL : O_TRUNC), 0644)) < 0) {
			fprintf(stderr, "could not open key file [%s] for writing: %s\n", keyfilename, strerror(errno));
			exit(2);
		}
	}

	for (i = 0; i < ANON_MAX_FIELDS; i++) {
		if (anon_keys[i].valid) {
			printf("sizeof(*anon_keys[%d].bf_ctx)): %d\n", i, sizeof(*anon_keys[i].bf_ctx));
			anon_keys[i].bf_ctx = malloc(sizeof(BF_KEY));
			if (!anon_keys[i].bf_ctx) {
				perror("malloc");
				exit(2);
			}
			BF_set_key(anon_keys[i].bf_ctx, BF_KEYSIZE, (unsigned char *)anon_keys[i].key);
		}
	}


	if ((outfd = open(outfilename, O_WRONLY | O_CREAT | (!clobber ? O_EXCL : O_TRUNC), 0644)) < 0) {
		fprintf(stderr, "could not open output file [%s] for writing: %s\n", outfilename, strerror(errno));
		exit(2);
	}
	total_bytes_anonymized = 0;
	err = parse(&params);

	fprintf(stderr, "Parsing OK!\n");

	if (err != ERR_OK)
		fprintf(stderr, "Parsing failed with error code %d\n", err);

	if (do_anonymize) {
		if (err == ERR_OK) err = dumpkeys(keyfilefd, BF_KEYSIZE);
		close(keyfilefd);
		if (err != ERR_OK)
			unlink(keyfilename);

	}

	for (i = 0; i < ANON_MAX_FIELDS; i++) {
		if (anon_keys[i].valid) {
			free(anon_keys[i].bf_ctx);
		}
	}

	if (params.bytes_read > 0) {
		printf("Total bytes anonymized = %d of %d, [%0.2f%%]\n", total_bytes_anonymized, params.bytes_read, (total_bytes_anonymized * 100.0 / params.bytes_read));
	}


	close(randomfd);
	close(outfd);
	return err;
}

struct descriptor {
	op_t type;
	char *str_type;
};

char *type_to_str_type(op_t type)
{
	int i;
	for (i = 0; msg_desc[i].str_type != NULL; i++) {
		if (type == msg_desc[i].msg_type)
			return msg_desc[i].str_type;
	}
	return "MSG_UNKNOWN";
}

void *on_start_trace_message(op_t type, int pre_or_post,void* context)
{
	struct buffer *buf = malloc(sizeof(struct buffer));
	if (buf) {
		memset(buf, 0, sizeof(*buf));
		buf->data = malloc(BUF_SIZE);
		if (!buf->data) {
			free(buf);
			buf = NULL;
		} else {
			buf->size = BUF_SIZE;
			buf->ptr = buf->data + INIT_LOC;
			memcpy(buf->data, &type, sizeof(type));
		}
	}
	return buf;
}
int on_end_trace_message(op_t type, int pre_or_post, void *cont)
{

    struct buffer *buf = cont;
    int err = ERR_OK;
    msg_len_t len = buf->ptr - buf->data - INIT_LOC;
    memcpy(buf->data + LEN_LOC, &len, sizeof(msg_len_t));

    if (write(outfd, buf->data, buf->ptr - buf->data) != (buf->ptr - buf->data))
	err = ERR_IO;
    free(buf->data);
    free(buf);
    return err;
}

void on_error(int error_type, int msg_num, char *desc)
{
	fprintf(stderr, "Error occurred while parsing trace: err_num = %d, msg_num = %d, description = %s\n", error_type, msg_num, desc);
}

void on_warning(int warn_type, char *desc)
{
    fprintf(stderr, "Warning: %s [%d]\n", desc, warn_type);
}

int on_data_field(op_t op, int pre_or_post, int tlv, field_t type, int length, void *value, void *cont)
{
	struct buffer *buf = cont;
	char *extra = NULL;
	int err = ERR_OK;
	field_len_t len = length;
	int space_needed;
	int i = ANON_MAX_FIELDS;
	if (do_anonymize) {
		switch (type) {
			case ARG_UID:
				if (anon_keys[ANON_UID].valid) i = ANON_UID;
				break;
			case ARG_GID:
				if (anon_keys[ANON_GID].valid) i = ANON_GID;
				break;
			case ARG_PID:
				if (anon_keys[ANON_PID].valid) i = ANON_PID;
				break;
			case ARG_PNAME:
				if (anon_keys[ANON_PNAME].valid) i = ANON_PNAME;
				break;
			case ARG_PATHNAME:
				if (anon_keys[ANON_PATHNAME].valid) i = ANON_PATHNAME;
				break;
			case ARG_STR:
				if (anon_keys[ANON_STR].valid) i = ANON_STR;
				break;
			case ARG_TIMESTAMP:
				if (anon_keys[ANON_TIMESTAMP].valid) i = ANON_TIMESTAMP;
				break;
		}
		if (i != ANON_MAX_FIELDS) {
			int padl;
			char iv[BF_BLOCKSIZE];

			memset(iv, '\0', BF_BLOCKSIZE);
			extra = malloc(len + BF_BLOCKSIZE);
			memcpy(extra, value, length);
			padl = BF_BLOCKSIZE - (length & (BF_BLOCKSIZE - 1));
			memset(extra + length, padl, padl);
			len = length + padl;
			BF_cbc_encrypt((unsigned char *)extra,(unsigned char *) extra, len, anon_keys[i].bf_ctx,(unsigned char *) iv, 1);

			total_bytes_anonymized += length;
			value = extra;

		}

	}
	space_needed = sizeof(field_t) + sizeof(field_len_t) + len;
	if (buf->data + buf->size - buf->ptr < space_needed) {
		// increase size by space_needed
		char *tmp = realloc(buf->data, buf->size + space_needed);
		if (!tmp) {
			err = ERR_NOMEM;
			goto out;
		} else {
			buf->ptr = tmp + (buf->ptr - buf->data);
			buf->data = tmp;
		}
	}
	if (i != ANON_MAX_FIELDS || tlv) {
		type |= FIELD_LENGTH_MASK;
		memcpy(buf->ptr, &type, sizeof(type));
		buf->ptr += sizeof(op);
		memcpy(buf->ptr, &len, sizeof(len));
		buf->ptr += sizeof(len);
		memcpy(buf->ptr, value, len);
		buf->ptr += len;
	} else {
		memcpy(buf->ptr, &type, sizeof(type));
		buf->ptr += sizeof(op);
		memcpy(buf->ptr, value, len);
		buf->ptr += len;
	}

out:
	if (extra)
		free(extra);
	return err;
}

int on_header(struct stream_header *header)
{
	if (write(outfd, (void *)header, sizeof(*header)) != sizeof(*header))
		return ERR_IO;
	else
		return ERR_OK;
}
