/*
 * Copyright (c) 2003-2006 Erez Zadok
 * Copyright (c) 2003-2006 Stony Brook University
 * Copyright (c) 2003-2006 The Research Foundation of State University of New York
 * Copyright (c) 2004-2006 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2006 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: dentry.c,v 1.21 2006/11/28 03:05:32 ezk Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_tracefs.h"
#endif /* FISTGEN */
#include "fist.h"
#include "tracefs.h"
#include "tracer.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
#undef d_delete

/*
 * THIS IS A BOOLEAN FUNCTION: returns 1 if valid, 0 otherwise.
 */
STATIC int
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
tracefs_d_revalidate(dentry_t *dentry, int flags)
#else
tracefs_d_revalidate(dentry_t *dentry, struct nameidata* nd)
#endif
{
    int err = 1; // default is valid (1); invalid is 0.
    dentry_t *hidden_dentry;
    void *context;

    print_entry_location();
    if (dentry->d_inode) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	trace_d_revalidate_pre(stoptr(dentry->d_inode->i_sb), dentry, flags, &context);
#else
	trace_d_revalidate_pre(stoptr(dentry->d_inode->i_sb), dentry, nd, &context);
#endif
    }
    hidden_dentry  = tracefs_hidden_dentry(dentry); /* CPW: Moved to after print_entry_location */

    if (!hidden_dentry->d_op || !hidden_dentry->d_op->d_revalidate)
	goto out;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    err = hidden_dentry->d_op->d_revalidate(hidden_dentry, flags);
#else
    err = hidden_dentry->d_op->d_revalidate(hidden_dentry, nd);
#endif
 out:
    if (dentry->d_inode) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	trace_d_revalidate_post(stoptr(dentry->d_inode->i_sb), dentry, flags, err, &context);
#else
	trace_d_revalidate_post(stoptr(dentry->d_inode->i_sb), dentry, nd, err, &context);
#endif
    }
    print_exit_status(err);
    return err;
}


STATIC int
tracefs_d_hash(dentry_t *dentry, qstr_t *name)
{
    int err = 0;
    dentry_t *hidden_dentry;
    void *context;

    print_entry_location();
    if (dentry->d_inode)
	trace_d_hash_pre(stoptr(dentry->d_inode->i_sb), dentry, name, &context);
    hidden_dentry = tracefs_hidden_dentry(dentry); /* CPW: Moved to after print_entry_location */

    if (!hidden_dentry->d_op || !hidden_dentry->d_op->d_hash || !hidden_dentry->d_op->d_hash)
	goto out;

    err = hidden_dentry->d_op->d_hash(hidden_dentry, name);

 out:
    if (dentry->d_inode)
	trace_d_hash_post(stoptr(dentry->d_inode->i_sb), dentry, name, err, &context);

    print_exit_status(err);
    return err;
}


STATIC int
tracefs_d_compare(dentry_t *dentry, qstr_t *a, qstr_t *b)
{
    int err;
    dentry_t *hidden_dentry;
    void *context;

    print_entry_location();
    if (dentry->d_inode)
	trace_d_compare_pre(stoptr(dentry->d_inode->i_sb), dentry, a, b, &context);

    hidden_dentry = tracefs_hidden_dentry(dentry); /* CPW: Moved to after print_entry_location */

    if (hidden_dentry && hidden_dentry->d_op && hidden_dentry->d_op->d_compare) {
	// XXX: WRONG: should encode_filename on a&b strings
	err = hidden_dentry->d_op->d_compare(hidden_dentry, a, b);
    } else {
	err = ((a->len != b->len) || memcmp(a->name, b->name, b->len));
    }

    if (dentry->d_inode)
	trace_d_compare_post(stoptr(dentry->d_inode->i_sb), dentry, a, b, err, &context);

    print_exit_status(err);
    return err;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
int
tracefs_d_delete(dentry_t *dentry)
{
    dentry_t *hidden_dentry = NULL;
    int err = 0;
    void *context;

    print_entry_location();

    if(!dentry)
	goto out;

    if (dentry->d_inode)
	trace_d_delete_pre(stoptr(dentry->d_inode->i_sb), dentry, &context);

    /* this could be a negative dentry, so check first */
    if (!dtopd(dentry)) {
	fist_dprint(6, "dentry without private data: %*s", dentry->d_name.len, dentry->d_name.name);
	goto out;
    }

    if (!(hidden_dentry = dtohd(dentry))) {
	fist_dprint(6, "dentry without hidden_dentry: %*s", dentry->d_name.len, dentry->d_name.name);
	goto out;
    }

    //    fist_print_dentry("D_DELETE IN", dentry);
    /* added b/c of changes to dput(): it calls d_drop on us */
    if (hidden_dentry && hidden_dentry->d_op &&
	hidden_dentry->d_op->d_delete) {
	err = hidden_dentry->d_op->d_delete(hidden_dentry);
    }

out:
    if (dentry->d_inode)
	trace_d_delete_post(stoptr(dentry->d_inode->i_sb), dentry, err, &context);

    print_exit_status(err);
    return err;
}
#endif

void
tracefs_d_release(dentry_t *dentry)
{
    dentry_t *hidden_dentry;
    void *context;

    print_entry_location();

    fist_print_dentry("tracefs_d_release IN dentry", dentry);

    trace_d_release_pre(stoptr(dentry->d_sb), dentry, &context);

    /* this could be a negative dentry, so check first */
    if (!dtopd(dentry)) {
	fist_dprint(6, "dentry without private data: %*s", dentry->d_name.len, dentry->d_name.name);
	goto out;
    }
    hidden_dentry = dtohd(dentry);
    fist_print_dentry("tracefs_d_release IN hidden_dentry", hidden_dentry);\
    if (hidden_dentry && hidden_dentry->d_inode)
	fist_dprint(6, "tracefs_d_release: hidden_inode->i_count %d, i_num %lu.\n",
		    atomic_read(&hidden_dentry->d_inode->i_count),
		    hidden_dentry->d_inode->i_ino);

    /* free private data (tracefs_dentry_info) here */
    KFREE(dtopd(dentry));
    //dtopdlh(dentry) = NULL;	/* just to be safe */
    /* decrement hidden dentry's counter and free its inode */
    if(hidden_dentry)
	dput(hidden_dentry);
 out:
    trace_d_release_post(stoptr(dentry->d_sb), dentry, &context);
    print_exit_location();
}


/*
 * we don't really need tracefs_d_iput, because dentry_iput will call iput() if
 * tracefs_d_iput is not defined. We left this implemented for ease of
 * tracing/debugging.
 */
void
tracefs_d_iput(dentry_t *dentry, inode_t *inode)
{
    void *context;
    print_entry_location();
    trace_d_iput_pre(stoptr(inode->i_sb), dentry, inode, &context);
    /*
     * tim: calling trace_d_iput_post() before iput(), since inode might
     * not exist after iput(), does not have any effect on the underlying
     * tracefs code in stream and aggregate traces.
     */
    trace_d_iput_post(stoptr(inode->i_sb), dentry, inode, &context);
    iput(inode);
    print_exit_location();
}


struct dentry_operations tracefs_dops = {
    d_revalidate:	tracefs_d_revalidate,
    d_hash:		tracefs_d_hash,
    d_compare:		tracefs_d_compare,
    d_release:		tracefs_d_release,
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
    d_delete:		tracefs_d_delete,
#endif
    d_iput:		tracefs_d_iput,
};


/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
